/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
#include "license.readme"

#include "myspSuggestmgr.h"
#include "plstr.h"
#include "nsCRT.h"
#include "nsReadableUtils.h"

myspSuggestMgr::myspSuggestMgr()
{
}


myspSuggestMgr::~myspSuggestMgr()
{
  pAMgr = nsnull;
  maxSug = 0;
}

void 
myspSuggestMgr::setup(const nsAFlatCString &tryme, int maxn, myspAffixMgr *aptr)
{
  // register affix manager and check in string of chars to 
  // try when building candidate suggestions
  pAMgr = aptr;
  ctry = tryme;
  maxSug = maxn;
}


// generate suggestions for a mispelled word
//    pass in address of array of char * pointers

nsresult myspSuggestMgr::suggest(char ***slst,const nsAFlatCString &word, PRUint32 *num)
{
  if(!num || !slst)
    return NS_ERROR_NULL_POINTER;
  nsresult res;
  PRUint32 nsug;
  PRUint32 i;
  char **wlst;
  if(!(*slst)){
    nsug=0;
    wlst=(char **)nsMemory::Alloc(sizeof(char *)*maxSug);
    if(!wlst)
      return NS_ERROR_OUT_OF_MEMORY;
    for(i=0;i<maxSug;i++)
      wlst[i]=nsnull;
  }
  else{
    wlst=*slst;
    nsug=*num;
  }

  // did we forget to add a char
  res = forgotchar(wlst, word, &nsug);

  // did we swap the order of chars by mistake
  if ((nsug < maxSug) && NS_SUCCEEDED(res)){
    res = swapchar(wlst, word, &nsug);
  }

  // did we add a char that should not be there
  if ((nsug < maxSug) && NS_SUCCEEDED(res)){
    res = extrachar(wlst, word, &nsug);
  }

    // did we just hit the wrong key in place of a good char
  if ((nsug < maxSug) && NS_SUCCEEDED(res)){
    res = badchar(wlst, word, &nsug);
  }

    // perhaps we forgot to hit space and two words ran together
  if ((nsug < maxSug) && NS_SUCCEEDED(res)){
    res = twowords(wlst, word, &nsug);
  }
  if(NS_FAILED(res)){
    for (i=0;i<maxSug; i++)
      if (wlst[i] != NULL) nsMemory::Free(wlst[i]);
    free(wlst);
    *slst = 0;
    *num=0;
  }
  else{
    *slst=wlst;
    *num=nsug;
  }
  return res;
}



// error is wrong char in place of correct one
nsresult myspSuggestMgr::badchar(char ** wlst,const nsAFlatCString &word, PRUint32 *ns)
{
  char	tmpc;
  nsSharableCString candidate;
  PRBool cwrd;
  PRUint32 i,j,k;
  PRUint32 wl = word.Length();
  candidate.Assign(word);
  nsACString::iterator candIt;
  for (i=0,candidate.BeginWriting(candIt); i < wl; i++,candIt++) {
    tmpc = *candIt;
    for (j=0; j < ctry.Length(); j++) {
      if (ctry[j] == tmpc) continue;
      *candIt = ctry[j];
      cwrd = PR_TRUE;
      for(k=0;k < *ns;k++){
        if (candidate.Equals(wlst[k]) ){ 
          cwrd = PR_FALSE;
          break;
        }
      }
      if (cwrd && pAMgr->check(candidate)) {
        if (*ns < maxSug) {
          wlst[*ns] = ToNewCString(candidate);
          if(!wlst[*ns])
            return NS_ERROR_OUT_OF_MEMORY;
          (*ns)++;
        } else return NS_OK;
      }
      *candIt = tmpc;
    }
  }
  return NS_OK;
}


// error is word has an extra letter it does not need 
nsresult myspSuggestMgr::extrachar(char ** wlst,const nsAFlatCString &word, PRUint32 *ns)
{
  PRBool cwrd;
  nsString stCand;
  nsSharableCString candidate;
  PRUint32 k;
  PRUint32 wl = word.Length();
  if (wl < 2) return 0;
  
  // try omitting one char of word at a time
  candidate.Assign(Substring(word,1,wl-1));
  nsACString::iterator r;
  nsACString::const_iterator p,end;
  word.EndReading(end);
  
  for (word.BeginReading(p),candidate.BeginWriting(r);  p != end;  ) {
    cwrd = PR_TRUE;
    for(k=0;k < *ns;k++){
      if (candidate.Equals(wlst[k])){ 
        cwrd = PR_FALSE;
        break;
      }
    }
    if (cwrd && pAMgr->check(candidate)) {
      if (*ns < maxSug) {
        wlst[*ns] = ToNewCString(candidate);
        if(!wlst[*ns])
          return NS_ERROR_OUT_OF_MEMORY;
        (*ns)++;
      } else return NS_OK;
    }
    *r++ = *p++;
  }
  return NS_OK;
}


// error is mising a letter it needs
nsresult myspSuggestMgr::forgotchar(char **wlst,const nsAFlatCString &word, PRUint32 *ns)
{
  PRBool cwrd;
  nsString stCand;
  nsSharableCString candidate;
  PRUint32 i,k;
  candidate.Assign(" ");
  candidate.Append(word);
  nsACString::iterator q;
  nsACString::const_iterator p,end;
  word.EndReading(end);

  // try inserting a tryme character before every letter
  for (word.BeginReading(p), candidate.BeginWriting(q);  p != end;  )  {
    for ( i = 0;  i < ctry.Length();  i++) {
      *q = ctry[i];
      cwrd = PR_TRUE;
      for(k=0;k < *ns;k++){
        if (candidate.Equals(wlst[k]) ){ 
          cwrd = PR_FALSE;
          break;
        }
      }
      if (cwrd && pAMgr->check(candidate)) {
        if (*ns < maxSug) {
          wlst[*ns] = ToNewCString(candidate);
          if(!wlst[*ns])
            return NS_ERROR_OUT_OF_MEMORY;
          (*ns)++;
        } else return NS_OK;
      }
    }
    *q++ = *p++;
  }
   
  // now try adding one to end */
  for ( i = 0;  i < ctry.Length();  i++) {
    *q = ctry[i];
    cwrd = PR_TRUE;
    for(k=0;k < *ns;k++){
      if (candidate.Equals(wlst[k])){ 
        cwrd = PR_FALSE;
        break;
      }
    }
    if (cwrd && pAMgr->check(candidate)) {
      if (*ns < maxSug) {
        wlst[*ns] = ToNewCString(candidate);
        if(!wlst[*ns])
          return NS_ERROR_OUT_OF_MEMORY;
        (*ns)++;
      } else return NS_OK;
    }
  }
  return NS_OK;
}


/* error is should have been two words */
nsresult myspSuggestMgr::twowords(char ** wlst,const nsAFlatCString &word, PRUint32 *ns)
{
  nsSharableCString candidate;
  nsString stCand;
  PRUint32 pos;
  PRUint32 wl=word.Length();
  if (wl < 3) return NS_OK;
  candidate.Assign(word);
  nsSharableCString temp;

  // split the string into two pieces after every char
  // if both pieces are good words make them a suggestion
  for (pos =  1;  pos < wl;  pos++) {
    temp.Assign(Substring(candidate,0,pos));
    if (pAMgr->check(temp)) {
      temp.Assign(Substring(candidate,pos,wl-pos));
      if (pAMgr->check(temp)) {
        if (*ns < maxSug) {
          candidate.Insert(' ',pos);
          wlst[*ns] = ToNewCString(candidate);
          if(!wlst[*ns])
            return NS_ERROR_OUT_OF_MEMORY;
          (*ns)++;
        } else return NS_OK;
      }
    }
  }
  return NS_OK;
}


// error is adjacent letter were swapped
nsresult myspSuggestMgr::swapchar(char **wlst,const nsAFlatCString &word, PRUint32 *ns)
{
  nsSharableCString candidate;
  char	tmpc;
  PRBool cwrd;
  nsString stCand;
  PRUint32 k;
  candidate.Assign(word);
  nsACString::iterator p,q,end;
  candidate.EndWriting(end);

  for (candidate.BeginWriting(p),q=p, q++;  q != end;  p++,q++) {
    tmpc = *p;
    *p = *q;
    *q = tmpc;
    cwrd = PR_TRUE;
    for(k=0;k < *ns;k++){
      if (candidate.Equals(wlst[k])){ 
        cwrd = PR_FALSE;
        break;
      }
    }
    if (cwrd && pAMgr->check(candidate)) {
      if (*ns < maxSug) {
        wlst[*ns] = ToNewCString(candidate);
        if(!wlst[*ns])
          return NS_ERROR_OUT_OF_MEMORY;
        (*ns)++;
      } else return NS_OK;
    }
    tmpc = *p;
    *p = *q;
    *q = tmpc;
  }
  return NS_OK;
}

